(*  Press 
    Copyright (C) 2005, 2006 M. Sighireanu and L. Allali

    This file is part of Press.

    Press is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Press is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Press; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*)

(*
 *  Implementation of the reachability algorithms.
 *
 *  File:         $RCSfile: prs_reach_0.ml,v $
 *  Author:       L. Allali and M. Sighireanu
 *  Version:      $Revision: 1.4 $
 *  Date:         $Date: 2006/05/31 19:44:34 $
 *)


(* ******************************************************************
 * Import of modules.
 * ******************************************************************)
open Symbol
open Prs_reach_util
open Prs_reach_eq

(* ******************************************************************
 * Exported function build_closure_0: Closure of an automaton by ~0.
 * ******************************************************************)

(** Constant needed.  
 *)
let str_qNull = "qNull";;

(** Build rules for idle neutrality.
 *  in build_rules_neutral_idle :
 *  0 -> qNull
 *  seq(qNull, qNull) -> qNull
 *  par(qNull, qNull) -> qNull
 *  par0(qNull, qNull) -> qNull
 *  in rule_ini_3_6_q :
 *  seq(qNull, q) -> q
 *  seq(q, qNull) -> q
 *  par(qNull, q) -> q
 *  par(q, qNull) -> q
 *  par0(qNull, q) -> q
 *  par0(q, qNull) -> q
 *)
let rec
    build_rules_neutral_idle a = 
	let str_rule_1 = (idleOp^" -> "^str_qNull)
	and str_rule_2_3 = 
	  (seqOp^"("^str_qNull^", "^str_qNull^") -> "^str_qNull^" "^
	   parOp^"("^str_qNull^", "^str_qNull^") -> "^str_qNull^" "^
	   par0Op^"("^str_qNull^", "^str_qNull^") -> "^str_qNull)
	and str_rule_4_7 = (String.concat " " (List.map rule_ini_4_7_q (States.to_list (Automaton.get_states a)))) 

	in
	(str_rule_1^" "^str_rule_2_3^" "^str_rule_4_7^" ")

and rule_ini_4_7_q q = (
	let str_q = (Terms.to_string q) in
		seqOp^"("^str_qNull^" , "^str_q^") -> "^str_q^"  "^
		seqOp^"("^str_q^" , "^str_qNull^") -> "^str_q^"  "^
	 	parOp^"("^str_qNull^" , "^str_q^") -> "^str_q^"  "^
		parOp^"("^str_q^" , "^str_qNull^") -> "^str_q^"  "^
	 	par0Op^"("^str_qNull^" , "^str_q^") -> "^str_q^"  "^
		par0Op^"("^str_q^" , "^str_qNull^") -> "^str_q
	)


(** Main entry for computation of closure by ~0. *)
let build_closure_0 a0 r =
	let sigma = (Automaton.get_alphabet a0) and
            sigmaq0 = (Automaton.get_state_ops a0) and
	    q0 = (Automaton.get_states a0) and
	    delta0 = (Automaton.get_transitions a0) in
	let aR = (build_aut_prs sigma r) in
        let qR = (Automaton.get_states aR) and
	    deltaR = (Automaton.get_transitions aR) and
	    sigmaqR = (Automaton.get_state_ops aR) in
        (** new alphabet for states *)
        let sigmaqN = (string2stateOps str_qNull) in
        let sigmaqNull = Sigma.union sigmaqN (Sigma.union sigmaq0 sigmaqR) in
	
	let qnullSet = (string2stateSet sigma sigmaqNull str_qNull) in
	(** new set of states *)
	let statesNull = States.union qnullSet (States.union q0 qR) in 

	(** new set of transitions *)
	let a = (Automaton.make_automaton
                       sigma
                       sigmaqNull
                       statesNull
                       (Automaton.get_final_states a0)
                       (PRS.empty)
                       (PRS.union delta0 deltaR)
                )
	in
	let trs = (string2trs sigma sigmaqNull (build_rules_neutral_idle a)) in 
	let aNull = (Automaton.make_automaton 
		       sigma 
		       sigmaqNull
		       statesNull
                       (Automaton.get_final_states a)
                       (PRS.empty)
                       (PRS.union (Automaton.get_transitions a) trs)
		    )
	in
	aNull
;;






(* ******************************************************************
 * Local functions for adding rules.
 * ******************************************************************)

(** Apply rule 1: 
 *  q -> qT 
 *  @see Prs_reach_eq.build_reach_1 a qSet 
 *)
let rec build_reach_0_rule1 aNull qSet=
  let a1 = (build_reach_1 aNull qSet) in
  (Automaton.get_transitions a1)
;;
  


(** Apply rule 2: 
 *  if t1->t2 and t1->*qT then qt2T->qT  
 *)
let rec build_reach_0_rule2 a r qSetT = 
  let rec
      rule2_r_s aRun srcRun srcBuild dst =
	(Prs_reach_util.if_run_then_build_trs aRun srcRun dst aRun [srcBuild] [dst]) 
  and
      rule2_r aRun qSet rule = (
      	let lhsTerm = (PRS.lhs rule) and
	    rhsTerm = (string2term 
				(Automaton.get_alphabet aRun)
                           	(Automaton.get_state_ops aRun)
                		((to_state_name (PRS.rhs rule))^suffixT))
	in
  	(List.fold_left PRS.union PRS.empty
		(List.map (rule2_r_s aRun lhsTerm rhsTerm) (States.to_list qSet)))
	)
  in
  let trs2 = (List.fold_left PRS.union PRS.empty 
		(List.map (rule2_r a qSetT) (PRS.to_list r)))
  in 
  let a2 = (Automaton.make_automaton
		(Automaton.get_alphabet a)
		(Automaton.get_state_ops a)
		(Automaton.get_states a)
		(Automaton.get_final_states a)
		(PRS.empty)
		trs2)
  in
  begin
  	(print_debug_aut a2 "rule2" true);
  	trs2
  end
;;
  

(** Apply rule 3 and rule 4: 
 *  3) if seq(q1,q2)->q in deltaNull then seq(q1T,q2)->qT  
 *  4) if par(q1,q2)->q in deltaNull then par(q1T,q2T)->qT 
 *  4') if par0(q1,q2)->q in deltaNull then par0(q1T,q2T)->qT 
 *)
let rec build_reach_0_rule3_4 a deltaNull=
  let sigma = (Automaton.get_alphabet a) and
      sigmap = (Automaton.get_state_ops a)
  in
  let rule3_4 rule = 
    let lhs = (PRS.lhs rule) and 
	rhs = (PRS.rhs rule) in
    let op = Symbol.to_string (Terms.top_symbol lhs) in
    
    (if (((op = seqOp) || (op = parOp) || (op = par0Op)) && 
	 (Terms.is_special rhs)) 
    then
	    let q1 = (List.hd (Terms.list_arguments lhs)) and
		q2 = (List.hd (List.tl (Terms.list_arguments lhs))) and
		q = (Terms.get_special rhs)
	    in  
	    let str_q1T = (Terms.to_string q1)^suffixT and
		str_q2T = (Terms.to_string q2)^suffixT and
		str_qT = (Terms.to_string q)^suffixT
	    in
	    let str_new = (
		if (op=seqOp) then 
	      	(seqOp^"("^str_q1T^","^(Terms.to_string q2)^") -> "^str_qT)
	    	else 
	      	(op^"("^str_q1T^","^str_q2T^") -> "^str_qT)
		)
	    in
	      (string2trs sigma sigmap (str_new))
    else 
	    PRS.empty)
      
  in
  let trs3 = (List.fold_left PRS.union PRS.empty 
			(List.map rule3_4 (PRS.to_list deltaNull)))
  in
  let a3 = (Automaton.make_automaton
		(Automaton.get_alphabet a)
		(Automaton.get_state_ops a)
		(Automaton.get_states a)
		(Automaton.get_final_states a)
		(PRS.empty)
		trs3)
  in
  begin
        (print_debug_aut a3 "rule3" true);
        trs3
  end
;;
  


(** Apply rule 5: 
 *	if q -> q' in deltaNull then qT->q'T 
 *  BUT  we have no rule such as q->q'
 *  x->q and q->qT is replaced by x->q and x->qT 
 *)
  

(** Apply rule 6: 
 *    if 0->*(a)q1T et seq(q1,q2)->q in deltaNull alors q2T->qT 
 *    if 0->*(a)q1T et par?(q1,q2)->q in deltaNull alors q2T->qT 
 *    if 0->*(a)q1T et par?(q2,q1)->q in deltaNull alors q2T->qT
 *)
let rec build_reach_0_rule6 a deltaNull =
	let trs6 = (List.fold_left PRS.union PRS.empty 
			(List.map (build_reach_0_rule6_r a) 
				  (PRS.to_list deltaNull))) in
	let a6 = (Automaton.make_automaton
			(Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        trs6)
       in
       begin
	(print_debug_aut a6 "rule6" true);
	trs6
       end

and build_reach_0_rule6_r a rule =
	let sigma = (Automaton.get_alphabet a) and
	    sigmap = (Automaton.get_state_ops a) and
	    lhs = (Automaton.lhs rule) and 
	    rhs = (Automaton.rhs rule) in
	let t0 = (string2term sigma sigmap idleOp)
	in
	let op = Symbol.to_string (Terms.top_symbol lhs)
  	in 
	(if (((op = seqOp) || (op = parOp) || (op = par0Op)) && 
	     (Terms.is_special rhs)) 
	then
		let q1_q2 = (Automaton.list_state rule) in
		let q1 = (List.hd q1_q2) and
		    q2 = (List.hd (List.tl q1_q2)) and
		    q = (Automaton.state_label rhs)
		in  
		let q1T = (to_state_T q1) and
		    q2T = (to_state_T q2) and
		    qT = (to_state_T q)
		in
		let rule6_1 = (
			if (op = seqOp) then
			    (Prs_reach_util.if_run_then_build_trs 
					a t0 q1T 
					a [(state2term sigma sigmap q2T)] [qT])
			else
			    PRS.empty
			)
                and rule6_2 = (
			if ((op = parOp) || (op = par0Op)) then
			    (PRS.union
			         (Prs_reach_util.if_run_then_build_trs 
					a t0 q1T 
					a [(state2term sigma sigmap q2T)] [qT])
			         (Prs_reach_util.if_run_then_build_trs 
					a t0 q2T 
					a [(state2term sigma sigmap q1T)] [qT]))
			else
			    PRS.empty
			)
		in
		   (PRS.union_fast rule6_1 rule6_2)
	  else 
		(PRS.empty) 
	 )
;;



(** Add rule 7: seq(qNullT,qT)->qT *)

let rec build_reach_0_rule7 aNull qSet=
	let sigma = (Automaton.get_alphabet aNull) and
	    sigmap = (Automaton.get_state_ops aNull) in
	
	let rule7 trs q = 
		let str_qT = (Terms.to_string q)^suffixT
		in
		let str_trs =seqOp^"("^str_qNull^suffixT^","^str_qT^") -> "^str_qT in
		let trsOk = (string2trs sigma sigmap str_trs)
		in 
    		    (PRS.union trs trsOk) (* TODO: speed up *)	
	in
	let trs7 = (List.fold_left rule7 
			(Automaton.get_transitions aNull)
			((States.to_list qSet)))
	in 
	let a7 = (Automaton.make_automaton
                        (Automaton.get_alphabet aNull)
                        (Automaton.get_state_ops aNull)
                        (Automaton.get_states aNull)
                        (Automaton.get_final_states aNull)
                        (PRS.empty)
                        trs7)
  	in
  	begin
        (print_debug_aut a7 "rule7" true);
        trs7
  	end
;;
  

(** Apply rules 1 and 7.
 *)
let build_reach_0_17 aNull qNull=
  let trs1 = (build_reach_0_rule1 aNull qNull)
  and trs7 = (build_reach_0_rule7 aNull qNull) in 
  let trs17 = (PRS.union (PRS.union (Automaton.get_transitions aNull) trs1) trs7) in
  (Automaton.make_automaton 
     (Automaton.get_alphabet aNull)  
     (Automaton.get_state_ops aNull) 
     (Automaton.get_states aNull)  
     (Automaton.get_final_states aNull)
     (PRS.empty) 
     trs17)
;;


(** Apply rules 2 to 6.
 *)
let build_reach_0_23456 a r qNull qNullT deltaNull =
  let trs2  = (build_reach_0_rule2 a r qNullT)
  and trs34 = (build_reach_0_rule3_4 a deltaNull)
  (* no need rule 5 *)
  and trs6  = (build_reach_0_rule6 a deltaNull) in 
  let trs23456 = (List.fold_left PRS.union (Automaton.get_transitions a) [trs2; trs34; trs6])
  in
  (Automaton.make_automaton 
     (Automaton.get_alphabet a)  
     (Automaton.get_state_ops a) 
     (Automaton.get_states a)  
     (Automaton.get_final_states a)
     (PRS.empty) 
     trs23456)
;;

(** Apply all rules of the algorithm in a fixed point computation.
 *)
let build_reach_0 a r qNull qNullT deltaNull = 
 
  (* rule 1 and 7 out of lfp *)
  let a17 = build_reach_0_17 a qNull in 
  
  let rec build_reach_0_loop a r qNull qNullT deltaNull =
  begin
      (print_new_iteration ());
      no_iteration := !no_iteration + 1 ;
      (let a23456 = (build_reach_0_23456 a r qNull qNullT deltaNull) in
       if (List.length (PRS.to_list (Automaton.get_transitions a))) = 
         (List.length (PRS.to_list (Automaton.get_transitions a23456)))
       then
	a 
       else 
	build_reach_0_loop a23456 r qNull qNullT deltaNull
      )
  end
  in 
  (build_reach_0_loop a17 r qNull qNullT deltaNull)
  
;;


(* ******************************************************************
 * Main algorithm.
 * ******************************************************************)

let build_aut_reach_0 a0 r = 
	begin
        (print_message "build reachability automaton modulo ~0");
        (let aNull = (build_closure_0 a0 r) in
	 let sigma = (Automaton.get_alphabet aNull) and
	     sigmaqNull = (Automaton.get_state_ops aNull) and
	     qNull = (Automaton.get_states aNull) and
	     qFNull = (Automaton.get_final_states aNull) and
	     deltaNull = (Automaton.get_transitions aNull) in
         
	 let str_sigmaqNullT = (build_states_reachT qNull)  and
             str_qFNullT = (build_states_reachT qFNull) in
         let sigmaqNullT = (string2stateOps str_sigmaqNullT) in
         let sigmaq = Sigma.union_fast sigmaqNull sigmaqNullT in
	 let qNullT = (string2stateSet sigma sigmaq str_sigmaqNullT) in
	 let qNullAndT= (States.union_disjoint qNull qNullT) in
	 let qFT = (States.union_disjoint qFNull (string2stateSet sigma sigmaq str_qFNullT)) in
	

	 let aNull_closed = (Automaton.make_automaton 
			sigma 
			sigmaq
			qNullAndT 
			qFT 
			(PRS.empty) 
			deltaNull) in
	 begin
         (print_debug_aut aNull_closed "aNull_star" false);
	 let lfp = (build_reach_0 aNull_closed r qNull qNullT deltaNull) 
	 and lfp_name = (!filename_debug)^"_clean"^".aut.tbk" in
         begin
         (print_debug_aut lfp "full" false);
         (print_message "clean the automaton");
         let lfp_red = (Automaton.accessibility_cleaning lfp) in
         begin
                (if (!option_equiv = 2) then
                        (print_message ("=====> see file "^lfp_name))
                 else ()
                )
		;
	 	(print_debug_aut lfp_red "clean" false);
                lfp_red
         end
         end
	 end
	)
	end
;;



