(*  Press 
    Copyright (C) 2005, 2006 M. Sighireanu and L. Allali

    This file is part of Press.

    Press is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Press is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Press; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*)

(*
 *  Implementation of the reachability algorithm modulo ~obar equivalence.
 *
 *  File:         $RCSfile: prs_reach_p.ml,v $
 *  Author:       M. Sighireanu
 *  Version:      $Revision: 1.3 $
 *  Date:         $Date: 2006/05/31 19:44:34 $
 *)


(* ******************************************************************
 * Import of modules.
 * ******************************************************************)

open Symbol
open Prs_reach_util
open Prs_reach_0
open Prs_reach_s


(* ******************************************************************
 * Local functions.
 * ******************************************************************)


(** Definitions for manipulation of parctx states.
 *)

(** Constant used.  *)
let str_parctx = "parctx"
;;

(** Build the name of state from q and X => parctx_q_x_.
 *)
let build_state_parctx l_r q var =
	if l_r = true then
        (str_parctx^"_"^var^"_"^(Terms.to_string q)^"_")
	else
	" "
	(* (str_parctx^"_"^(Terms.to_string q)^"_"^var^"_") *)
;;


(** Build the states corresponding to parctx (q,X) and parctx(X,q).  *)
let build_states_parctx sigma qSet =
        let build_states_parctx_q_var q var = (
                if (Prs_reach_util.is_user_var var) then
                begin
                        (* (print_message var); *)
                        (build_state_parctx false q var)^" "^(build_state_parctx true q var)
                end
                else
                        ""
                )
        in
        let build_states_parctx_q sigma q =
                (String.concat " " (List.map (build_states_parctx_q_var q) (Sigma.to_string_list sigma)))
        in
        let str_states_parctx = (String.concat " " (List.map (build_states_parctx_q sigma) (States.to_list qSet)))
        in
                str_states_parctx
;;


(** Definitions used by the algorithm.
 *)


(** Algorithm rule alpha2:
 *  reach_s: if Y ->*(a) q and seq(X,Y) -> t in R then qtT -> (q,X).
 *  if Y ->*(a) q and par0(X,Y) -> t in R then qtT -> |X,q|
 *  if X ->*(a) q and par0(X,Y) -> t in R then qtT -> |q,Y|
 *)
let rec
    build_reach_p_2 a r qNull = 
        let trs2 = (List.fold_left PRS.union PRS.empty
                        (List.map (build_reach_p_2_r a qNull) (PRS.to_list r)))
        in
        let a2 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs2))
        in
        begin
                (print_debug (function () ->
                        let a2_name = "it"^(string_of_int !no_iteration)^"_rule2" in
                        let a2_filename = (!filename_debug)^"_"^a2_name^".aut.tbk" in
                        (SpecPRS.save_automaton a2 a2_name a2_filename)));
                a2
        end

and build_reach_p_2_r a qNull r =
    let op = (Symbol.to_string (Terms.top_symbol (PRS.lhs r))) in
	(if (op = seqOp) || (op = par0Op) then
	    let lhs = (PRS.lhs r) and
		rhs = (PRS.rhs r) in
            (* assert: PRS in normal form *)
            let varX = (List.hd (Terms.list_arguments lhs)) and
                varY = (List.hd (List.tl (Terms.list_arguments lhs))) in
            let rhsT = (string2term
                         (Automaton.get_alphabet a)
                         (Automaton.get_state_ops a)
                         ((to_state_name rhs)^suffixT)) in
		(if (op = seqOp) then
		 (List.fold_left PRS.union PRS.empty
                        (List.map (Prs_reach_s.build_reach_s_2_r_q a varY rhsT varX) (States.to_list qNull)))
		else (* op = par0Op *)
		 (List.fold_left PRS.union PRS.empty
                        (List.map (build_reach_p_2_r_q a varY rhsT varX) (States.to_list qNull)))
		)
	else
		PRS.empty
	)

and build_reach_p_2_r_q a tY tT tX q =
	let qy = (string2state (build_state_parctx false q (Terms.to_string tY)))
	and xq = (string2state (build_state_parctx true q (Terms.to_string tX)))
	in
	(* PRS.union_fast  *)
		(Prs_reach_util.if_run_then_build_trs a tY q a [tT] [xq])
	(*	(Prs_reach_util.if_run_then_build_trs a tX q a [tT] [qy])
	 *)
;;


(** Algorithm rule alpha3:
 *  if exists t1->t2 in r and q in qNull s.t.:
 *  a)-b) (@see Prs_reach_s.build_reach_s_3)
 *  c) t1 ->*(a) |q,X| then qt2T -> |q,X|
 *  d) t1 ->*(a) |X,q| then qt2T -> |X,q|
 *)
let build_reach_p_3 a r qNullT qSeqCtx qParCtx = 
  	let trs3 = (List.fold_left PRS.union PRS.empty
                	(List.map (Prs_reach_s.build_reach_3_r a 
				(States.union_disjoint qNullT
				(States.union_disjoint qSeqCtx qParCtx))) 
				  (PRS.to_list r)))
	in
	let a3 = (Automaton.make_automaton
	                 (Automaton.get_alphabet a)
       		         (Automaton.get_state_ops a)
       		         (Automaton.get_states a)
       		         (Automaton.get_final_states a)
       		         (PRS.empty)
       		         (PRS.union (Automaton.get_transitions a) trs3))
    	in
        begin
        (print_debug_aut a3 "rule3" true);
        a3
        end
;;


(** Algorithm rules alpha4 + alpha5:
 *  It is like reach_s with more rules for 5 (@see # rules).
 *
 *  forall op(q1, q2) -> q in deltaNull (for 4 op = seqOp, for 5 op = par0Op)
 *  4a) seq(q1T, q2) -> qT                5a) par0(q1T,q2T) -> qT
 *      @see Prs_reach_0.build_reach_0_rule3_4
 *  4b) forall X seq(q1X, q2) -> qX	  5d#) par0(xq1, q2) -> xq
 *                                             par0(q1, q2X) -> qX
 *  4c) forall X if 0 ->*(a) q1X 	  5e#) if 0 ->*(a) |X,q1| 
 *               then q2T -> qX 	       then q2T -> |X,q|
 *                                             if 0 ->*(a) |q2,X|
 *                                             then q1T -> |q,X|
 *  4d) if 0 ->*(a) q1T then q2T -> qT    5b) same
 *                                        5c) if 0 ->*(a) q2T then q1T -> qT
 *      @see Prs_reach_0.build_reach_0_rule6
 *  4e) forall X . if X ->*(a) q1T        5f#) if X ->*(a) q1T
 *                 then q2X -> qT              then |X,q2| -> qT
 *                                             if X ->*(a) q2T
 *                                             then |q1,X| -> qT
 *  4f) forall X, X' . if X ->*(a) q1X'   5g#) if X ->*(a) |X',q1|
 *                     then q2X -> qX'         then |X,q2| -> |X',q|
 *                                             if X ->*(a) |q2, X'|
 *                                             then |q1,X| -> |q,X'|
 *)
let rec
    build_reach_p_4_5 a qNullT qSeqCtx qParCtx deltaNull = 
        let trs45 = List.fold_left PRS.union PRS.empty
                        (List.map (build_reach_p_4_5_r a qNullT qSeqCtx qParCtx)
                                  (PRS.to_list deltaNull))
        in
        let a45 = (Automaton.make_automaton
                        (Automaton.get_alphabet a)
                        (Automaton.get_state_ops a)
                        (Automaton.get_states a)
                        (Automaton.get_final_states a)
                        (PRS.empty)
                        (PRS.union (Automaton.get_transitions a) trs45))
        in
        begin
        (print_debug_aut a45 "rule45" true);
        a45
        end

and build_reach_p_4_5_r a qNull qSeqCtx qParCtx rule =
    let sigma = (Automaton.get_alphabet a) and
        sigmap = (Automaton.get_state_ops a) in
    let lhs = (PRS.lhs rule) and
        rhs = (PRS.rhs rule) in
    let op = Symbol.to_string (Terms.top_symbol lhs) and
        t0 = (string2term sigma sigmap idleOp)
    in

    (if (((op = seqOp) || (op = parOp) || (op = par0Op)) &&
         (Terms.is_special rhs))
    then
        let q1 = (List.hd (Terms.list_arguments lhs)) and
            q2 = (List.hd (List.tl (Terms.list_arguments lhs))) and
             q = (Terms.get_special rhs)
        in
        let str_q1 = (Terms.to_string q1) and
            str_q2 = (Terms.to_string q2) and
            str_q  = (Terms.to_string q)
        in
        let str_q1T = str_q1^suffixT and
            str_q2T = str_q2^suffixT and
            str_qT  = str_q^suffixT
        in
        let q1T = (string2term sigma sigmap str_q1) and
            q2T = (string2term sigma sigmap str_q2) and
             qT = (string2term sigma sigmap str_q)
        in
	let trs_4a = (Prs_reach_s.build_reach_s_4a sigma sigmap op str_q1 str_q2 str_q)
        and trs_5a = (Prs_reach_s.build_reach_s_5a sigma sigmap op str_q1 str_q2 str_q)
	and trs_4d = (Prs_reach_s.build_reach_s_4d a op t0 str_q1T q2T str_qT)
	and trs_5bc = (Prs_reach_s.build_reach_s_5bc a op t0 str_q1T str_q2T q2T q1T str_qT)
	and trs_45_X = (List.fold_left PRS.union PRS.empty
                       (List.map 
			(build_reach_p_45_r_X a sigma sigmap op t0 q1 q1T q2 q2T q qT)
                        (List.filter Prs_reach_util.is_user_var
                                        (Sigma.to_string_list sigma))
                       )
        		)
	in
	(List.fold_left PRS.union PRS.empty
		[trs_4a; trs_5a; trs_4d; trs_5bc; trs_45_X])
     else
        PRS.empty
     )

(* 5d) forall X . par0(|X,q1|, q2) -> |X,q|
 *                par0(q1, |q2,X|) -> |q,X|
 *)
and build_reach_p_5d_r_X sigma sigmap op q1 str_xq1 q2 str_q2x str_xq str_qx =
	(* assert op = par0Op *)
	(string2trs sigma sigmap
		(op^"("^str_xq1^","^(Terms.to_string q2)^") -> "^str_xq)) (*^" "^
                 op^"("^(Terms.to_string q1)^","^str_q2x^") -> "^str_qx)) *)

(*  5e#) forall X . if 0 ->*(a) |X,q1| then q2T -> |X,q|
 *                  if 0 ->*(a) |q2,X| then q1T -> |q,X|
 *)
and build_reach_p_5e_r_X a op t0 str_xq1 q1T str_q2x q2T str_xq str_qx =
	(* assert op = par0Op *)
	(let trs_5e1 = (Prs_reach_util.if_run_then_build_trs
				a t0 (string2state str_xq1)
				a [q2T] [(string2state str_xq)])
         (* and trs_5e2 = (Prs_reach_util.if_run_then_build_trs
				a t0 (string2state str_q2x)
				a [q1T] [(string2state str_qx)]) *)
	 in (* PRS.union_fast *) trs_5e1 (* trs_5e2 *)
	)
	

(*  5f#) forall X . if X ->*(a) q1T then |X,q2| -> qT
 *                  if X ->*(a) q2T then |q1,X| -> qT
 *)
and build_reach_p_5f_r_X a sigma sigmap op tX str_q1T str_q1x str_q2T str_xq2 str_qT = 
	(* assert op = par0Op *)
	(let trs_5f1 = (Prs_reach_util.if_run_then_build_trs
				a tX (string2state str_q1T)
				a [(string2term sigma sigmap str_xq2)] [(string2state str_qT)])
         (* and trs_5f2 = (Prs_reach_util.if_run_then_build_trs
				a tX (string2state str_q2T)
				a [(string2term sigma sigmap str_q1x)] [(string2state str_qT)]) *)
	 in (* PRS.union_fast *) trs_5f1 (* trs_5f2 *)
	)

(*  5g#) forall X, X' . if X ->*(a) |X',q1| then |X,q2| -> |X',q|
 *                      if X ->*(a) |q2, X'| then |q1,X| -> |q,X'|
 *)
and build_reach_p_5g_r_X_Xp a sigma sigmap op tX q1 str_q1x q2 str_xq2 q vX vXp = 
	(* assert op = par0Op *)
	(let str_xpq1 = (build_state_parctx true q1 vXp)
         and str_q2xp = (build_state_parctx false q2 vXp)
         and str_xpq = (build_state_parctx true q vXp)
	 and str_qxp = (build_state_parctx false q vXp)
	 in 
	 let trs_5g1 = (Prs_reach_util.if_run_then_build_trs
				a tX (string2state str_xpq1)
				a [(string2term sigma sigmap str_xq2)] [(string2state str_xpq)])
	 (* and trs_5g2 = (Prs_reach_util.if_run_then_build_trs
				a tX (string2state str_q2xp)
				a [(string2term sigma sigmap str_q1x)] [(string2state str_qxp)]) *)
	 in
		(* PRS.union_fast *) trs_5g1 (* trs_5g2 *)
	)


and build_reach_p_45_r_X a sigma sigmap op t0 q1 q1T q2 q2T q qT vX =
	(if (op = seqOp) then
	  (Prs_reach_s.build_reach_s_4bcef_X a sigma sigmap op t0 q1 q1T q2 q2T q qT vX)
 
         else if (op = par0Op) then
	   let str_xq1 = (build_state_parctx true q1 vX)
           and str_q1x = (build_state_parctx false q1 vX)
           and str_q2x = (build_state_parctx false q2 vX)
           and str_xq2 = (build_state_parctx true q2 vX)
           and str_xq = (build_state_parctx true q vX)
	   and str_qx = (build_state_parctx false q vX)
           and tX = (string2term sigma sigmap vX)
	   in
	   let trs_5d = (build_reach_p_5d_r_X sigma sigmap op q1 str_xq1 q2 str_q2x str_xq str_qx)
	   and trs_5e = (build_reach_p_5e_r_X a op t0 str_xq1 q1T str_q2x q2T str_xq str_qx)
	   and trs_5f = (build_reach_p_5f_r_X a sigma sigmap op tX (Terms.to_string q1T) str_q1x (Terms.to_string q2T) str_xq2 (Terms.to_string qT))
	   and trs_5g = (List.fold_left PRS.union PRS.empty
                           (List.map
                              (build_reach_p_5g_r_X_Xp a sigma sigmap op tX q1 str_q1x q2 str_xq2 q vX)
                              (List.filter Prs_reach_util.is_user_var
                                 (Sigma.to_string_list sigma))
                           )
			)
	   in
	   (List.fold_left PRS.union PRS.empty [trs_5d;trs_5e;trs_5f;trs_5g])

	 else (* op = parOp *)
	   PRS.empty
	)
;;


(** Algorithm rule alpha6:
 *)
let build_reach_p_6 a qNull deltaNull =
        (* since deltaNull is normalized,
         * there are no rules q -> q' in deltaNull *)
        a
;;


(** Build reach automaton by applying rules alpha1-7.
 *  @param a    initially is automaton with (Q~, Sigma, F~, deltaNull).
 *  @param r    prs
 *)
let rec
    build_reach_p a r qNull qNullT qSeqCtx qParCtx deltaNull =
        (* rules alpha1, alpha6, and alpha7 do not need iteration *)
        let a1 = Prs_reach_s.build_reach_s_1 a qNull in
        let a16 = build_reach_p_6 a1 qNull deltaNull in
        let a167 = Prs_reach_s.build_reach_s_7 a16 qNull in
        build_reach_p_2345 a167 r qNull qNullT qSeqCtx qParCtx deltaNull

and build_reach_p_2345 a r qNull qNullT qSeqCtx qParCtx deltaNull =
        begin
        (print_new_iteration ());
        no_iteration := !no_iteration + 1;
        (let a2 = (build_reach_p_2 a r qNull) in
         let a3 = (build_reach_p_3 a2 r qNullT qSeqCtx qParCtx) in
         let a45 = (build_reach_p_4_5 a3 qNullT qSeqCtx qParCtx deltaNull) in
         if (List.length (PRS.to_list (Automaton.get_transitions a))) =
            (List.length (PRS.to_list (Automaton.get_transitions a45))) then
                a
         else
                (build_reach_p_2345 a45 r qNull qNullT qSeqCtx qParCtx deltaNull)
        )
        end
;;


(* ******************************************************************
 * Main algorithm.
 * ******************************************************************)

let build_aut_reach_p a0 r = 
	begin
        (print_message "build reachability automaton modulo ~|");
        (let aNull = (build_closure_0 a0 r) in
	 let sigma = (Automaton.get_alphabet aNull) and
	     sigmaqNull = (Automaton.get_state_ops aNull) and
	     qNull = (Automaton.get_states aNull) and
	     qFNull = (Automaton.get_final_states aNull) and
	     deltaNull = (Automaton.get_transitions aNull) in
	 let str_sigmaqNullT = (build_states_reachT qNull) in
	 let str_qFNullT = (build_states_reachT qFNull) in
         let sigmaqNullT = (string2stateOps str_sigmaqNullT) in
	 let str_sigmaqSeqCtx = (Prs_reach_s.build_states_seqctx sigma qNull) in
	 let sigmaqSeqCtx = (string2stateOps str_sigmaqSeqCtx) in
	 let str_sigmaqParCtx = (build_states_parctx sigma qNull) in
	 let sigmaqParCtx = (string2stateOps str_sigmaqParCtx) in
         let sigmaq = (Sigma.union_fast sigmaqNull 
			(Sigma.union_fast sigmaqNullT 
			(Sigma.union_fast sigmaqSeqCtx sigmaqParCtx))) in
	 let qNullT = (string2stateSet sigma sigmaq str_sigmaqNullT) in
	 let qSeqCtx = (string2stateSet sigma sigmaq str_sigmaqSeqCtx) in
	 let qParCtx = (string2stateSet sigma sigmaq str_sigmaqParCtx) in
	 let qT = (States.union_disjoint qNull 
			(States.union_disjoint qNullT
			(States.union_disjoint qSeqCtx qParCtx))) in
	 let qFT = (States.union_disjoint qFNull (string2stateSet sigma sigmaq str_qFNullT)) in
	 let reachp = (Automaton.make_automaton 
			sigma 
			sigmaq
			qT 
			qFT 
			(PRS.empty) 
			deltaNull) in
	 begin
         (print_debug_aut reachp "init" false);
	 let lfp = (build_reach_p reachp r qNull qNullT qSeqCtx qParCtx deltaNull) and
	     lfp_name = (!filename_debug)^"_clean.aut.tbk" in
	 begin
         (print_debug_aut lfp "full" false);
	 (print_message "clean the automaton");
	 let lfp_red = (Automaton.clean lfp) in
         begin
                (if (!option_equiv = 3) then
                        (print_message ("=====> see file "^lfp_name))
                 else ()
                )
		;
         	(print_debug_aut lfp_red "clean" false);
                lfp_red
	 end
         end
	 end
	)
	end
;;

(* End of $RCSfile: prs_reach_p.ml,v $ *)
