(*  Press 
    Copyright (C) 2005, 2006 M. Sighireanu and L. Allali

    This file is part of Press.

    Press is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Press is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Press; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*)

(*
 *  Implementation of common utility functions.
 *
 *  File:         $RCSfile: prs_reach_util.ml,v $
 *  Author:       M. Sighireanu
 *  Version:      $Revision: 1.5 $
 *  Date:         $Date: 2006/05/31 19:44:34 $
 *)

(* ******************************************************************
 * Timbuk specific work.
 * ******************************************************************)

(** Import of Timbuk modules.
 *)
open Symbol
open Alphabet
open Variable
open Variable_set
open Term
open State_set
open State_content
open Rewrite
open Automaton
open Gamma
open Specification
open Completion

(** Module instanciation.
 *)
module Sigma= Alphabet(Symbol)
module Vars= Variable_set(Variable)
module Terms= Term(Symbol)(Sigma)(Variable)(Vars)
module PRS= RewriteSystem(Sigma)(Vars)(Terms)
module States= State_set(Symbol)(Sigma)(Terms)(State_content)
module Automaton= TreeAutomata(Symbol)(Sigma)(Variable)(Terms)(State_content)(PRS)(States)
module Gamma= Gamma(Symbol)(Sigma)(Variable)(Vars)(Terms)(PRS)(States)(Automaton)
module SpecPRS= Specification(Sigma)(Vars)(Terms)(PRS)(Automaton)(Gamma)

(* Useful functions *)
let string2state s = (Automaton.make_state (Symbol.of_string s))
;;
let string2stateOps s = (States.parse_ops (SpecPRS.lexer (Stream.of_string s)))
;;
let string2stateSet a o s = (States.parse a o (SpecPRS.lexer (Stream.of_string s)))
;;

let string2term a o s = (Terms.check_special a o (Terms.parse_ground_special a o (SpecPRS.lexer (Stream.of_string s))))
;;

let string2trs a o s = (PRS.check_special (PRS.parse_ground_special a o (SpecPRS.lexer (Stream.of_string s))) a o)
;;

let state2term a o s = (Terms.check_special a o (Common.Special s))
;;

let term2state t = (Automaton.make_state_config (Automaton.make_state (Terms.top_symbol (Terms.get_special t))))
;;

(* configuration of the garbage collector for lowest memory usage *)

let new_gc_control= {Gc.minor_heap_size=32768; Gc.major_heap_increment=63488;
 Gc.space_overhead=42; Gc.verbose=0; Gc.max_overhead=10;
 Gc.stack_limit=262144};;
Gc.set new_gc_control;;


(* ******************************************************************
 * Useful functions not provided by Timbuk.
 * ******************************************************************)
(** Empty automaton used to return from errors.
 *)
let automaton_empty = 
	(Automaton.make_automaton 
		(Sigma.new_alphabet) 
		(Sigma.new_alphabet) 
		(States.empty)
		(States.empty) 
		(PRS.empty) 
		(PRS.empty)
	) 
;;

let rec
(** Build the list of subterms for the term t and list lt. *)
    get_subterms lt t =
	let lt1 = lt@(Terms.list_arguments t) in
	    (if (List.length lt1) == 0 then [t]
	     else let newlt = (get_subterms (List.tl lt1) (List.hd lt1)) in
		term_list_union newlt [t] 
	    )
and
(** Disjoint union of two list of terms. *)
    term_list_union l1 l2 = 
	(match l1, l2 with
	  [], [] -> []
	| t1::l11, [] -> l1
	| [], t2::l21 -> l2
	| _, t2::l21  -> 
		(if (List.exists (Terms.equal t2) l1) then 
			(term_list_union l1 l21)
		 else
			[t2]@(term_list_union l1 l21)
		)
	)
;;

(* ******************************************************************
 * Messages to the user.
 * ******************************************************************)

(** Error handling and messages.
 *)
exception Error;;

type error_type =
	FILE_INPUT of string
    |   ALPHABET_ARITY of string
    |   ALPHABET_SYMBOL of string
    |   PRS_INPUT of string
    |   AUTOMATA_INPUT of string
;;

let print_error e = 
	begin
          print_string "Error ";
	  (match e with
	    FILE_INPUT s -> print_string ("(main): input file not given\n")
	  | ALPHABET_ARITY s -> print_string ("(parsing): incorrect arity for "^s^"\n")
	  | ALPHABET_SYMBOL s -> print_string ("(parsing): undefined operator "^s^"\n")
	  | PRS_INPUT s -> print_string ("(checking): PRS specification is "^s^"\n")
	  | AUTOMATA_INPUT s -> print_string ("(checking): automaton "^s^"\n")
	  );
	  raise Error;
	  ()
	end
;;

(** Log messages. *)
let print_message s = 
	begin
	(print_string ("\nprs_reach: - "^s));
	(if s = "exit." then (print_string "\n"));
	(Format.print_flush ())
	end
;;

(** Debug messages. *)
let no_iteration = ref 0
;;

let option_equiv = ref 0 ;;

let option_debug = ref false ;;

let filename_debug = ref "" ;;
let print_debug f = 
	(if (!option_debug) then
		(f ())
	 else
		()
	)
;;

let print_info_aut a a_name =
	begin
	(print_message ("\t automaton "^a_name^" "));
	(print_int (List.length (States.to_list (Automaton.get_states a))));
	(print_string " states, ");
	(print_int (List.length (PRS.to_list (Automaton.get_transitions a))));
	(print_string " transitions\n");
	(Format.print_flush ())
	end
;;

let print_debug_aut a msg is_it =
	let a_name = (if is_it = true then ("it"^(string_of_int !no_iteration)^"_") else "")^msg in
        let a_filename = (!filename_debug)^"_"^a_name^".aut.tbk" in
	begin
	(print_info_aut a a_name);
	(if (is_it = true) then
		(print_debug 
			(function () -> 
               		(SpecPRS.save_automaton a a_name a_filename)
			)
		)
	else
	(SpecPRS.save_automaton a a_name a_filename)
	)
	end
;;

let print_new_iteration () =
	begin (print_string "."); (Format.print_flush ()) end
;;

(* ******************************************************************
 * Checks on specification.
 * ******************************************************************)

(** Constants needed for mandatory operations.
 *)

let idleOp = "idle";;
let parOp = "par";;
let par0Op = "par0";;
let seqOp = "seq";;

(** Check that the alphabet contains the symbols for
 *    parallel composition (par), 
 *    sequential composition (seq), and
 *    idle process (idle).
 *  Exit if this is not the case.
 *)
let check_alphabet a =
	begin
        (print_message "check alphabet") ;
	(try
	  (if (Sigma.get_arity (Symbol.of_string parOp) a) != 2 then
		(print_error (ALPHABET_ARITY parOp))
	   else if (Sigma.get_arity (Symbol.of_string par0Op) a) != 2 then
                (print_error (ALPHABET_ARITY seqOp))
	   else if (Sigma.get_arity (Symbol.of_string seqOp) a) != 2 then
                (print_error (ALPHABET_ARITY seqOp))
	   else if (Sigma.get_arity (Symbol.of_string idleOp) a) != 0 then
                (print_error (ALPHABET_ARITY idleOp))
	   else 
		()
          )
         with Sigma.Symbol_not_in_alphabet s -> 
		print_error (ALPHABET_SYMBOL s)
	)
	end
;;


(** Check that the PRS is in normal form and return its transitions.
 *  A PRS is in normal form iff for each rule t1 -> t2 t1 and t2 have
 *  one of the forms idle, X, par(X,Y), seq(X,Y) with X,Y constants.
 *  
 *)
let rec 
    get_prs ltrs = 
	(match ltrs with
	  [] -> begin print_error (PRS_INPUT "empty") ; PRS.empty end
      	| t::l -> if (check_prs t) then (snd t) 
               	  else begin
          	  (print_error (PRS_INPUT "not in normal form")) ; 
	  	  PRS.empty
		  end
	)

and check_prs prs = 
	begin
	(print_message "check prs for normal form");
	(List.fold_left check_rule true (PRS.to_list (snd prs)))
	end

and check_rule b r = 
	(if not(b) then b
         else 
	 let res = (check_term 0 true (PRS.lhs r)) && 
		   (check_term 0 true (PRS.rhs r))
	 in begin
	    (if not res then
		(print_error (PRS_INPUT ("rule "^(PRS.rule_to_string r)^" not in normal form")))
	    );
	    res
            end
	)

and check_term depth b t = 
	(if not(b) then b
	 else
	(match t with
      	   Common.Fsym(symbol, l_subterms) ->
		let op = (Symbol.to_string symbol) in
        	(if (depth = 0) &&
           	    ((op = seqOp) ||
		     (op = parOp) || (op = par0Op)) then
		(List.fold_left (check_term 1) true l_subterms)
	 	else if (List.length l_subterms) = 0 then
			true
		else 
			false
		)
    	| Common.Const(_) -> true
    	| Common.Special(_) 
    	| Common.Var(_) -> false
	) 
	)
;;

(** Check that the initial automaton is given (name A0) and return it.
 *)
let get_init_aut laut = 
	begin
	(print_message "check initial automaton");
	try
	  List.assoc "init" laut
	with _ -> begin (print_error (AUTOMATA_INPUT "init empty")) ; automaton_empty end
	end
;;

(** Return true if it is a symbol different from par, seq, idle.
 *)
let is_user_var var = (
	if (var = idleOp) || (var = parOp) || (var = seqOp) || (var = par0Op)
        then false
	else true
	)
;;
	


(* ******************************************************************
 * Computation used by the construction of reachability automata.
 * ******************************************************************)

(* ******************************************************************
 * Building states.
 * ******************************************************************)

(** Constants needed for name of states.
 *)
let suffixT = "T";;
let charT = 'T';;
let suffixNil = "nil";;
let suffixNull = "null";;

(** Computation of the set of states.
 *)
let rec
(** Build the name of a state corresponding to a term. *)
   to_state_name term = 
	("q" ^ (to_state_name_simple term))
and
   to_state_name_simple term = 
	(match term with
	   Common.Fsym(symbol, l_subterms) ->
		let st = List.map to_state_name_simple l_subterms in
        		String.concat "" [(Symbol.to_string symbol); "_";
					  String.concat "_" st; "_"]
	 | Common.Const(c) -> (Symbol.to_string c)
	 | Common.Special(t) -> (to_state_name_simple t)
	 | Common.Var(x) -> (Variable.to_string x)
	)
;;

(** Build the string for the set of states derived from the right terms of prs.  
 *)
let build_states_prs lt = 
	let s = (String.concat " " (List.map to_state_name lt)) in
	begin
		(* (print_message ("\t Q_R = {"^s^"}")) *)
		s
	end

;;

(** Build the string for the set of states of reach automaton.
 *  This is done by adding T and NIL suffixes to each state in q. 
 *)
let rec
    build_states_reach str_q q = (
	let ql = States.to_list q in
	let s = (String.concat " " (List.map Terms.to_string ql))^" "^
		(build_states_reachT q)^" "^
		(build_states_reachNil q) in
	begin
	(* print_message ("\t "^str_q^" = {"^s^"}") *)
	s
	end
	)

and build_states_reachT q = (
	(String.concat (suffixT^"  ") 
		(List.map Terms.to_string (States.to_list q)))^suffixT
	)

and build_states_reachNil q = (
	(String.concat (suffixNil^" ") 
		(List.map Terms.to_string (States.to_list q)))^suffixNil 
	)
;;

(** Compute the derived states and terms (qT and qnil) from a state q.
 *)
let to_state_T q = 
	(string2state ((Terms.to_string q)^suffixT))
;;

let to_term_T sigma sigmap q = 
	(string2term sigma sigmap ((Terms.to_string q)^suffixT))
;;

let to_state_nil q = 
	(string2state ((Terms.to_string q)^suffixNil))
;;

let to_term_nil sigma sigmap q = 
	(string2term sigma sigmap ((Terms.to_string q)^suffixNil))
;;


(** Return the list of q from qSet such that t ->*(a) q.
 *)
let rec
    get_succ_of_term a t qSet = (
	get_succ_of_term_aux a t (States.to_list qSet) [])

and get_succ_of_term_aux a t qList acc = (
	match qList with
	  [] -> acc
	| q::ql -> (
		if (Automaton.run t q a) then 
			(get_succ_of_term_aux a t ql (q::acc)) 
		else
			(get_succ_of_term_aux a t ql acc) 
		)
	)
;;

(* ******************************************************************
 * Building transitions.
 * ******************************************************************)

(** Build transitions associated with the right terms of prs.  *)
let rec
    build_delta_prs sigma sigmap sub_r_R = 
	let str_trs = (String.concat "\n" (List.map to_trs_from_term sub_r_R)) in
	let deltaR  = (string2trs sigma sigmap str_trs) in
	begin 
		(* print_message ("\t delta_R = {\n"^(PRS.to_string deltaR)^"\n}") *)
		deltaR
	end
and
(** Build the string t -> qt. *)
    to_trs_from_term t = (** assert: prs in normal form *) 
	(match t with
      Common.Fsym(symbol, l_subterms) ->
        let st = List.map to_state_name l_subterms in
        (Symbol.to_string symbol)^"("^(String.concat "," st)^")"
    | Common.Const(c) -> (Symbol.to_string c)
    | Common.Special(t) -> (to_state_name_simple t)
    | Common.Var(x) -> (Variable.to_string x)
	)^" -> "^(to_state_name t)
;;

(** Conditionnaly building of transitions.
 *
 *  Build the transition srcBuild -> dstBuild (eventually normalized in aBuild)
 *  if exists a run in aRun between srcRun and dstRun (srcRun ->*(aRun) dstRun)
 *
 *  @param aRun		automaton where the run is searched
 *  @param srcRun	term source of the run
 *  @param dstRun	state destination of the run
 *  @param aBuild	automaton where the transition build will be added
 *  @param srcBuild	list of term source of the transition
 *  @param dstBuild	list of state source of the transition
 *  @return 		normalized trs (set of transitions) in aBuild
 *)
let rec
    if_run_then_build_trs aRun srcRun dstRun aBuild srcBuild dstBuild =
	(* check that parameters are correct *)
	if (Automaton.run srcRun dstRun aRun) then
		(List.fold_left PRS.union PRS.empty
			(List.map2 
				(if_run_then_build_trs_aux aBuild) 
				srcBuild dstBuild))
        else
		PRS.empty

and if_run_then_build_trs_aux aBuild src dst =
	let sigma = (Automaton.get_alphabet aBuild) and
	    sigmaq = (Automaton.get_state_ops aBuild) in
	if (Automaton.is_state_config src) then
	(* transition should be normalized *)
		(Automaton.normalize_epsilon 
			src (state2term sigma sigmaq dst)
			(Automaton.get_transitions aBuild))
	else
		(string2trs sigma sigmaq ((Terms.to_string src)^" -> "^(Terms.to_string dst)))
(* (PRS.check_special (PRS.list_to_trs [(PRS.new_rule src (state2term sigma sigmaq dst))]) sigma sigmaq) *)
;;

(* ******************************************************************
 * Building automata.
 * ******************************************************************)

(** Build the automaton from the PRS rules.
 *)
let build_aut_prs sigma r =
	let terms_r_R = (List.map PRS.rhs (PRS.to_list r)) in
	let sub_r_R = (List.fold_left term_list_union [] (List.map (get_subterms []) terms_r_R)) in
	let str_sigmaR = (build_states_prs sub_r_R) in
	let sigmaR = (string2stateOps str_sigmaR) in
	let qR = (string2stateSet sigma sigmaR str_sigmaR) in
	let deltaR = (build_delta_prs sigma sigmaR sub_r_R) in
		(Automaton.make_automaton sigma sigmaR qR States.empty (PRS.empty) deltaR)
;;



